--------------------------------------------------------------------------------
--  File     : _helpers.lua
--  Author(s): Pavel Demchuk, Artyom Vorobets
--
--  Summary  : This is QE++, dude!
--
--  Copyright  2013-2015 Hullabu Inc.
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
--  
--------------------------------------------------------------------------------
function split(str, delimiter)
	local res = {}
	local pos = 1
	local startPos, endPos
	repeat
		startPos, endPos = str:find(delimiter, pos, true)
		if startPos then
			table.insert(res, str:sub(pos, startPos - 1))
			pos = endPos + 1
		else
			table.insert(res, str:sub(pos, str:len()))
		end
	until not startPos
	return res
end

--------------------------------------------------------------------------------
function join(pieces, delimiter)
	local res = ''
	for i = 1,(#pieces-1) do
		res = res .. pieces[i] .. delimiter
	end
	return (res .. pieces[#pieces])
end

--------------------------------------------------------------------------------
function join_fmt(fmt, from, to, delimiter)
	local res = ''
	for i = from,(to-1) do
		res = res .. string.format(fmt, i) .. delimiter
	end
	return (res .. string.format(fmt, to))
end

--------------------------------------------------------------------------------
function set(list)
	local res = {}
	for _,l in ipairs(list) do
		res[l] = true
	end
	return res
end

--------------------------------------------------------------------------------
function clamp(value, min, max)
	if value < min then return min end
	if value > max then return max end
	return value
end

--------------------------------------------------------------------------------
function shuffle_table(tbl)
	local n, order, res = #tbl, {}, {}
	for i = 1,n do
		order[i] = { rnd = math.random(), idx = i }
	end
	table.sort(order, function(a,b) return a.rnd < b.rnd end)
	for i = 1,n do
		res[i] = tbl[order[i].idx]
	end
	return res
end



--------------------------------------------------------------------------------
--       QuestEngine
--------------------------------------------------------------------------------
-- 
--------------------------------------------------------------------------------
ALPHA0	= ctypes:Color(0x00ffffff)
ALPHA1	= ctypes:Color(0xffffffff)

--------------------------------------------------------------------------------
--   print,      ae2d
--------------------------------------------------------------------------------
function print(s)
	LogError(tostring(s))
end

--------------------------------------------------------------------------------
--    cstore    
--------------------------------------------------------------------------------
function SetVar(name, value)
	cstore:SetKey(name, value)
end

--------------------------------------------------------------------------------
--     cstore  0
--------------------------------------------------------------------------------
function GetVar(name)
	return cstore:GetKey(name) or 0
end

--------------------------------------------------------------------------------
--     cstore  1    
--------------------------------------------------------------------------------
function IncVar(name)
	local res = GetVar(name) + 1
	SetVar(name, res)
	return res
end

--------------------------------------------------------------------------------
--    
--------------------------------------------------------------------------------
function ShowItem(name)
	room:GetItem(name):SetVisible(true)
end

--------------------------------------------------------------------------------
--    
--------------------------------------------------------------------------------
function HideItem(name)
	room:GetItem(name):SetVisible(false)
end

--------------------------------------------------------------------------------
--      
--------------------------------------------------------------------------------
function ShowHideItem(name, bShow)
	room:GetItem(name):SetVisible(bShow)
end

--------------------------------------------------------------------------------
--      
--------------------------------------------------------------------------------
function IsItemVisible(name)
	return room:GetItem(name):IsVisible()
end

--------------------------------------------------------------------------------
-- ,        
--------------------------------------------------------------------------------
function IsItemOnCursor(name)
	return cursor:IsItemNameEqual(name)
end

--------------------------------------------------------------------------------
--      ,     ","
--------------------------------------------------------------------------------
function ShowItems(list)
	for _,i in ipairs(split(list, ',')) do
		room:GetItem(i):Show()
	end
end

--------------------------------------------------------------------------------
function HideItems(list)
	for _,i in ipairs(split(list, ',')) do
		room:GetItem(i):Hide()
	end
end

--------------------------------------------------------------------------------
function LockItems(list)
	for _,i in ipairs(split(list, ',')) do
		room:GetItem(i):Disable()
	end
end

--------------------------------------------------------------------------------
function UnlockItems(list)
	for _,i in ipairs(split(list, ',')) do
		room:GetItem(i):Enable()
	end
end

--------------------------------------------------------------------------------
function ShowMagicStars(list, starsType)
	for _,i in ipairs(split(list, ',')) do
		stars:Show(i, starsType)
	end
end

--------------------------------------------------------------------------------
function IsInInventoryOneOf(list)
	for _,i in ipairs(split(list, ',')) do
		if inventory:IsItemExisting(i) then
			return true
		end
	end
end

--------------------------------------------------------------------------------
function SafeAddInventoryItems(list)
	for _,i in ipairs(split(list, ',')) do
		if not inventory:IsItemExisting(i) then
			inventory:AddInventoryItem(quest_items:GetInvItem(i))
		end
	end
end

--------------------------------------------------------------------------------
function SafeRemoveInventoryItems(list)
	for _,i in ipairs(split(list, ',')) do
		if inventory:IsItemExisting(i) then
			inventory:EraseOneItemFromCell(i)
		end
	end
end

--------------------------------------------------------------------------------
function GetCursorAngle(item, lastMousePos)
	local d = lastMousePos - (item:GetPosition() + item:GetRotCenter())
	return -math.deg(math.atan2(d.y, d.x))
end



--------------------------------------------------------------------------------
--      
--------------------------------------------------------------------------------
function SafeCallFunction(func, arg)
	if func then
		return func(arg)
	end
end

--------------------------------------------------------------------------------
function CloseClose(dialog)
	dialog:Close()
	dialog:Close()
end

--------------------------------------------------------------------------------
--      C++    PersSpeak
--------------------------------------------------------------------------------
function ShowSpeakDialog(tag, image, startSound, endSound)
	local pers_table = ui.PersSpeak:GetPersTable()
	local line = 0
	local base, full

	repeat
		base = tag .. string.format("_%04d_", line)
		for i = 1,#pers_table do
			full = base .. pers_table[i]
			if resm:IsStringExist(full, room:GetName()) then
				ui.PersSpeak:Add(full)
				line = line + 10
				break
			end
			full = nil
		end
	until not full

	--  ,      
	if line > 0 then
		local fade_opacity = 0.6
		if image then
			ui.PersSpeak:SetImage(image)
		elseif chapter:IsInCloseup() or resm:IsStringExist("PUZZLE_DESC", room:GetName()) then
			fade_opacity = 0
		end

		--    ,       fade_opacity  BlockPersSpeak
		for i = 1,(startSound and 2 or 1) do
			ui.ModalDialogsShadow:SetFillColorAlpha(fade_opacity)
		end
		ui.PersSpeak:AddSounds(startSound or "", endSound or "")
		ui.PersSpeak:SetUserTag(tag)
		ui.PersSpeak:Show()
		return true
	else
		--print("ShowSpeakDialog('" .. tag .. "') empty!")
	end
end

--------------------------------------------------------------------------------
--     
-- ..    ShowComment("Foo"),       Foo_Alex  Foo_Rhianna,
--    ,    
--------------------------------------------------------------------------------
function ShowComment(textIdBase, showOnce)
	local fullId
	for _,persName in ipairs(ui.PersSpeak:GetPersTable()) do
		fullId = textIdBase .. "_" .. persName
		if resm:IsStringExist(fullId, room:GetName()) then
			if not (showOnce and cstore["comm_" .. fullId]) then
				comment:Add(fullId)
				cstore["comm_" .. fullId] = true
			end
			return true
		end
	end
end

--------------------------------------------------------------------------------
--    ,   
--------------------------------------------------------------------------------
function Blink_Start(itemName, count, showMinTime, showMaxTime, hideMinTime, hideMaxTime, endFunc)
	if not BlinkData then
		BlinkData = {}
	end
	local t =
	{
		showMinTime		= showMinTime,
		showMaxTime		= showMaxTime,
		hideMinTime		= hideMinTime,
		hideMaxTime		= hideMaxTime,
		count			= count*2 + 1,
		timer			= 0,
		endFunc			= endFunc,
	}
	if BlinkData[itemName] then
		t.visibleOnStart = BlinkData[itemName].visibleOnStart
	else
		t.visibleOnStart = IsItemVisible(itemName)
	end
	BlinkData[itemName] = t
end

--------------------------------------------------------------------------------
function Blink_Stop(itemName)
	if BlinkData and BlinkData[itemName] then
		ShowHideItem(itemName, BlinkData[itemName].visibleOnStart)
		BlinkData[itemName] = nil
	end
end

--------------------------------------------------------------------------------
function Blink_Update(elapsedTime)
	if not BlinkData then
		return
	end

	local dtbl, item = {}, nil

	for itemName,params in pairs(BlinkData) do
		item = room:GetItem(itemName)
		params.timer = params.timer - elapsedTime
		if params.timer <= 0 then
			if item:IsVisible() then
				item:Hide()
				params.timer = params.hideMinTime + (params.hideMaxTime - params.hideMinTime) * math.random()
			else
				item:Show()
				params.timer = params.showMinTime + (params.showMaxTime - params.showMinTime) * math.random()
			end
			if params.count > 0 then
				params.count = params.count - 1
				if params.count == 0 then
					if params.endFunc then
						params.endFunc()
					end
					table.insert(dtbl, itemName)
				end
			end
		end
	end

	for _,itemName in ipairs(dtbl) do
		Blink_Stop(itemName)
	end
end

--------------------------------------------------------------------------------
--   
--------------------------------------------------------------------------------
function TextTyping_Start(itemName, stringBase, minDelay, maxDelay, typingSound, doneSound, endFunc)
	if not TextTypingData then
		TextTypingData = {}
	end
	TextTypingData[itemName] =
	{
		stringBase	= stringBase,
		minDelay	= minDelay,
		maxDelay	= maxDelay,
		typingSound	= typingSound,
		doneSound	= doneSound,
		letterIndex	= 1,
		timer		= 0,
		endFunc		= endFunc,
	}
end

--------------------------------------------------------------------------------
function TextTyping_Stop(itemName)
	if TextTypingData then
		TextTypingData[itemName] = nil
	end
end

--------------------------------------------------------------------------------
function TextTyping_Update(elapsedTime)
	if not TextTypingData then
		return
	end

	local dtbl = {}

	for itemName,params in pairs(TextTypingData) do
		params.timer = params.timer - elapsedTime
		if params.timer <= 0 then
			if resm:IsStringExist(params.stringBase .. params.letterIndex, room:GetName()) then
				if params.typingSound then
					room.sfx:Play(params.typingSound)
				end
				room:GetItem(itemName):SetTextId(params.stringBase .. params.letterIndex)
				params.letterIndex = params.letterIndex + 1
				params.timer = params.minDelay + (params.maxDelay - params.minDelay) * math.random()
			else
				if params.doneSound then
					room.sfx:Play(params.doneSound)
				end
				if params.endFunc then
					params.endFunc()
				end
				table.insert(dtbl, itemName)
			end
		end
	end

	for _,itemName in ipairs(dtbl) do
		TextTypingData[itemName] = nil
	end
end

--------------------------------------------------------------------------------
--   -
--------------------------------------------------------------------------------
function ShowVideo(name, fadeIn, fadeOut, actionOnEnd)
	--     ,      
	music:SetPause(true, false)
	ui.VideoPlayer:SetUserTag(actionOnEnd or name)
	ui.VideoPlayer:PlayVideo(name, fadeIn, fadeOut)
end

--------------------------------------------------------------------------------
--  -
--------------------------------------------------------------------------------
function OnDialogClose_VideoPlayer(dialog)
	local tag = dialog:GetUserTag()
	--       -     -      
	if tag ~= "show dashboard" then
		music:SetPause(false, true)
		SafeCallFunction(OnCloseVideoPlayer, tag)
	end
end

--------------------------------------------------------------------------------
--      .
--      primary_tasks:bla_bla:GetBlaBla():IsBlaBla()
--------------------------------------------------------------------------------
function NewTask(taskId)
	--     
	game_info:CreateTextEvent(taskId)

	local t = tasks_settings[taskId]
	local taskType = t[1]

	if taskType == "click" then
		-- CLICK (CreateSimple) - c  ,       OnStartTask(taskName)
		-- :
		-- [2]  ( nil,   icon_default)
		-- [3]   ( nil,   Start)
		primary_tasks:CreateSimple(taskId, t[2] or "icon_default", t[3] or "")
	elseif taskType == "pay" then
		-- PAY (CreatePaidSimple) -  ,         ,      OnStartTask(taskName)
		-- :
		-- [2] 
		-- [3]  
		-- [4]  ( nil,   icon_pay)
		-- [5]   ( nil,   Pay)
		primary_tasks:CreatePaidSimple(taskId, t[4] or "icon_pay", t[5] or "", ToCurrency(t[3]), t[2])
	elseif taskType == "wait" then
		-- WAIT (CreateTimeWait) - ,     ,     
		-- :
		-- [2]    
		-- [3]    
		-- [4]  ( nil,   icon_wait_time)
		-- [5]   ( nil,   Start)
		primary_tasks:CreateTimeWait(taskId, t[4] or "icon_wait_time", t[5] or "", t[2], t[3])
	elseif taskType == "visit" then
		-- VISIT (CreateJumpRoom) -    ,          
		-- :
		-- [2]     
		-- [3]  
		-- [4]  ,   ,  nil
		-- [5] ID ,   ,  nil
		-- [6]  ( nil,   icon_find_object)
		-- [7]   ( nil,   Visit    Start  -)
		local roomName = t[3]
		local roomCard
		local taskUnlockCost
		if dashboard:IsRoomCardExists(roomName) then
			roomCard = dashboard:GetRoomCard(roomName)
			dashboard:OpenRoomCard(roomName)
			if not roomCard:IsLocked() then
				taskUnlockCost = t[2]
			end
		elseif rooms_settings[roomName] then
			taskUnlockCost = rooms_settings[roomName].UnlockCost
		end

		if taskUnlockCost and taskUnlockCost > 0 then
			primary_tasks:CreatePaidJumpRoom(taskId, t[6] or "icon_find_object", t[7] or "", Currency_Stars, taskUnlockCost, roomName, t[4] or "", t[5] or "")
		else
			if roomCard then
				if t[4] then
					roomCard:AssignItem(t[4], 0)
				end
				if t[5] then
					roomCard:AssignDialog(t[5], 0)
				end
			end
			primary_tasks:CreateJumpRoom(taskId, t[6] or "icon_find_object", t[7] or "", roomName)
		end
	else
		print("NewTask - unknown type '" .. taskType .. "'")
	end
end

--------------------------------------------------------------------------------
function DelTask(taskId)
	primary_tasks:EraseTask(taskId)
	--     ,    
	if roomCard then
		roomCard:ClearAssignedDialog()
	end
end

--------------------------------------------------------------------------------
function IsTaskExists(taskId)
	return primary_tasks:IsTaskExists(taskId)
end

--------------------------------------------------------------------------------
function IsTaskUnlocked(taskId)
	return primary_tasks:IsTaskUnlocked(taskId)
end

--------------------------------------------------------------------------------
function ReplaceTask(oldTaskId, newTaskId)
	if IsTaskExists(oldTaskId) then
		DelTask(oldTaskId)
		NewTask(newTaskId)
		return true
	end
end

--------------------------------------------------------------------------------
--   Rate Now,      
--------------------------------------------------------------------------------
function ShowRateDialog(delay)
	--    ,     ,   
	delay = math.floor(delay * 1000)
	if delay ~= gstore.RATE_NOW then
		gstore.RATE_NOW = delay
	else
		gstore.RATE_NOW = delay + 1
	end
end

--------------------------------------------------------------------------------
--   Dashboard
--------------------------------------------------------------------------------
function JumpDashboard()
	chapter:JumpRoom(chapter_settings.DashboardRoom)
end

--------------------------------------------------------------------------------
--    
--------------------------------------------------------------------------------
function JumpStoryRoom(name)
   -- - 
	qe:SaveGameToProfile()
	ui.DashboardPanel:Close()
	top_panel:Close()
	chapter:JumpRoom(name)
end

--------------------------------------------------------------------------------
--      
--------------------------------------------------------------------------------
function UpdateDailyReward()
	local yday = os.date("*t").yday
	local diff = game_info.daily_reward_last_yday and (yday - game_info.daily_reward_last_yday) or -1
	if diff ~= 0 then
		game_info.daily_reward_last_yday = yday
		if diff == 1 or diff == -365 or diff == -366 then
			game_info.daily_reward_days_count = math.abs(game_info.daily_reward_days_count) + 1
		elseif diff > 0 then
			game_info.daily_reward_days_count = 1
		else
			game_info.daily_reward_days_count = 0
		end
	end
end

--------------------------------------------------------------------------------
function ShowDailyReward()
	if not ui:IsModal() then
		if game_info.daily_reward_days_count > 0 then
			ui.DailyReward:Show(game_info.daily_reward_days_count)
			game_info.daily_reward_days_count = -game_info.daily_reward_days_count
		elseif game_info.user_gift_json and game_info.user_gift_json ~= "" then
			ui.WeAreSorry:Show()
		end
	end
end

--------------------------------------------------------------------------------
-- ,      
--------------------------------------------------------------------------------
function CheckTimeForEvent(last_time, control_time_in_min)
	return (last_time == 0 or ((os.time() - last_time) / 60) >= control_time_in_min)
end

--------------------------------------------------------------------------------
--           '0'
--------------------------------------------------------------------------------
function GetCaseNumber(name)
	return tostring(tonumber((name or chapter:GetName()):sub(-2)))
end

--------------------------------------------------------------------------------
--      .
--        "rooms_settings"  .
--------------------------------------------------------------------------------
--   ,       
--------------------------------------------------------------------------------
function GetLevelXpPoints(level)
	if level <= 1 then return 0 end
	return (game_settings.LevelsXpPoints[level - 1] or game_settings.LevelsXpPoints[#game_settings.LevelsXpPoints])
end

--------------------------------------------------------------------------------
--     ,      
--------------------------------------------------------------------------------
function GetReward(name)
	local t = rooms_settings and rooms_settings[name]
	return (t and unlock_reward_profiles[t.UnlockRewardProfile or "Default"] or rewards[name])
end

--------------------------------------------------------------------------------
--   
--------------------------------------------------------------------------------
function TransferRewardToUser(reward, kontagentEventName)
	for currency,amount in pairs(reward) do
		if game_info[currency] then
			if currency == "energy" then
				--    ,    
				game_info.energy_bonus = game_info.energy_bonus + amount
			else
				game_info[currency] = game_info[currency] + amount
			end

			if type(kontagentEventName) == 'string' then
				kontagent:Event(kontagentEventName, amount, "economy", currency, "source")
			end
		else
			print("TransferRewardToUser - unknown currency '" .. currency .. "'")
		end
	end
end

--------------------------------------------------------------------------------
--      
--------------------------------------------------------------------------------
function GetBonusProfile(roomName)
	return bonus_profiles[rooms_settings[roomName].BonusProfile or "Default"]
end

--------------------------------------------------------------------------------
--      
--------------------------------------------------------------------------------
function GetRoomUnlockCost(roomName)
	return (rooms_settings[roomName].UnlockCost or 0)
end

--------------------------------------------------------------------------------
--  ,      
--------------------------------------------------------------------------------
function GetRoomUnlockTime(roomName)
	return (rooms_settings[roomName].UnlockTime or 0)
end

--------------------------------------------------------------------------------
--    -    
--------------------------------------------------------------------------------
function GetRoomSkipCost(roomName)
	return (rooms_settings[roomName].SkipCost or ((GetRoomUnlockTime(roomName) > 0) and game_settings.TimeLockSkipCost or game_settings.PuzzleSkipCost))
end

--------------------------------------------------------------------------------
--       
--------------------------------------------------------------------------------
function GetRoomStarTask(roomName, starNum)
	return stars_profiles[rooms_settings[roomName].StarsProfile][starNum]
end

--------------------------------------------------------------------------------
--     ,     
--------------------------------------------------------------------------------
function GetRoomPlayTime(roomName, starNum)
	local t = stars_profiles[rooms_settings[roomName].StarsProfile]
	return (t[starNum] and t[starNum].Time or (t.TimeWhenAllStarsDone or game_settings.TimeWhenAllStarsDone)) * game_settings.PlayTimeMultiplier
end

--------------------------------------------------------------------------------
--       
--------------------------------------------------------------------------------
function GetRoomGoldCupTime(roomName)
	local t = stars_profiles[rooms_settings[roomName].StarsProfile]
	return (t.GoldCupTime or game_settings.GoldCupTime) * game_settings.PlayTimeMultiplier
end

--------------------------------------------------------------------------------
--      
--------------------------------------------------------------------------------
function GetPlayEnergyCost()
	return (chapter_settings.PlayEnergyCost or game_settings.PlayEnergyCost)
end

--------------------------------------------------------------------------------
--       
--------------------------------------------------------------------------------
function CalcRoomBonusTimeScore(roomCard)
	local nextStar = math.min(roomCard:GetStarsCount() + 1, roomCard:GetMaxStarsCount())
	local starTask = GetRoomStarTask(roomCard:GetName(), nextStar)

	--   
	local step_size = 2
	--    0,2,4,6   ,      
	local scores = { 1000, 900, 800, 500, 400, 300, 200, 100 }
	--     
	local play_time = (top_panel.room_max_timer - top_panel.room_timer) / game_settings.PlayTimeMultiplier
	--      
	play_time = play_time / (starTask.TasksCount or game_settings.TasksCount)

	local a,b = math.modf(play_time / step_size)
	if (a+1) >= #scores or a < 0 then
		return scores[#scores]
	else
		return math.floor(scores[a+1] + (scores[a+2] - scores[a+1]) * b)
	end
end

--------------------------------------------------------------------------------
--        
--------------------------------------------------------------------------------
function CalcCollectibles()
--[[
	#tasks_settings				--    
	#collections[chapter:GetName()]	--  
	roomCard:GetCounterPasses()		--     
]]
	local rooms_list = {}
	local roomCard
	local allStarsEarned = true

	--   ,    
	for name,_ in pairs(rooms_settings) do
		if dashboard:IsRoomCardExists(name) then
			roomCard = dashboard:GetRoomCard(name)
			if roomCard:GetStarsCount() < roomCard:GetMaxStarsCount() then
				allStarsEarned = false
			end
			table.insert(rooms_list, name)
		end
	end

	--        
	local frags_left = ui.Collections:GetAvailableFragmentsCount(chapter:GetName())

	--    ,      
	if frags_left == 0 then
		for _,name in ipairs(rooms_list) do
			dashboard:GetRoomCard(name):ClearAssignedCollectible()
		end
		return
	end

	--       ,
	--         
	if allStarsEarned then
		for _,name in ipairs(rooms_list) do
			dashboard:GetRoomCard(name):AssignCollectible()
		end
	else
		math.randomseed(os.time())
		for _,name in ipairs(rooms_list) do
			roomCard = dashboard:GetRoomCard(name)
			--  ,           
			if not roomCard:IsCollectibleAssigned() and roomCard:GetStarsCount() < roomCard:GetMaxStarsCount() then
				local probability = 0.8 * (roomCard:GetStarsCount() + roomCard:GetNextStarProgress()) / roomCard:GetMaxStarsCount()
				--   ,      100%
				if game_info.case_progress[chapter:GetName()] == 100 or math.random() < probability then
					roomCard:AssignCollectible()
				end
			end
		end
	end
end
